# Allow classes to use self-referencing Type hints in Python 3.7.
from __future__ import annotations

import ray
import sys
from typing import Dict, Any
from dataclasses import dataclass
from deltacat import logs
import logging
from resource import getrusage, RUSAGE_SELF
import platform
import psutil


logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


@dataclass
class ClusterUtilization:
    def __init__(
        self, cluster_resources: Dict[str, Any], available_resources: Dict[str, Any]
    ):
        used_resources = {}

        for key in cluster_resources:
            if (
                isinstance(cluster_resources[key], float)
                or isinstance(cluster_resources[key], int)
            ) and key in available_resources:
                used_resources[key] = cluster_resources[key] - available_resources[key]

        self.total_memory_bytes = cluster_resources.get("memory")
        self.used_memory_bytes = used_resources.get("memory")
        self.total_cpu = cluster_resources.get("CPU")
        self.used_cpu = used_resources.get("CPU")
        self.total_object_store_memory_bytes = cluster_resources.get(
            "object_store_memory"
        )
        self.used_object_store_memory_bytes = used_resources.get("object_store_memory")
        self.used_memory_percent = (
            self.used_memory_bytes / self.total_memory_bytes
        ) * 100
        self.used_object_store_memory_percent = (
            self.used_object_store_memory_bytes / self.total_object_store_memory_bytes
        ) * 100
        self.used_cpu_percent = (self.used_cpu / self.total_cpu) * 100
        self.used_resources = used_resources

    @staticmethod
    def get_current_cluster_utilization() -> ClusterUtilization:
        cluster_resources = ray.cluster_resources()
        available_resources = ray.available_resources()

        return ClusterUtilization(
            cluster_resources=cluster_resources, available_resources=available_resources
        )


def get_current_node_peak_memory_usage_in_bytes():
    """
    Returns the peak memory usage of the node in bytes. This method works across
    Windows, Darwin and Linux platforms.
    """
    current_platform = platform.system()
    if current_platform != "Windows":
        usage = getrusage(RUSAGE_SELF).ru_maxrss
        if current_platform == "Linux":
            usage = usage * 1024
        return usage
    else:
        return psutil.Process().memory_info().peak_wset


def get_size_of_object_in_bytes(obj: object) -> float:
    size = sys.getsizeof(obj)
    if isinstance(obj, dict):
        return (
            size
            + sum(map(get_size_of_object_in_bytes, obj.keys()))
            + sum(map(get_size_of_object_in_bytes, obj.values()))
        )
    if isinstance(obj, (list, tuple, set, frozenset)):
        return size + sum(map(get_size_of_object_in_bytes, obj))
    return size



from __future__ import annotations

import copy
import json
from typing import Any, Dict, List, Optional

from deltacat.types.media import ContentType


class CompactPartitionParams(dict):
    """
    This class represents the parameters passed to compact_partition (deltacat/compute/compactor/compaction_session.py)
    """

    @staticmethod
    def of(params: Optional[Dict]) -> CompactPartitionParams:
        if params is None:
            params = {}
        compact_partition_params = CompactPartitionParams()
        compact_partition_params["destination_partition_locator"] = params.get(
            "destination_partition_locator"
        )
        compact_partition_params["last_stream_position_to_compact"] = params.get(
            "last_stream_position_to_compact"
        )
        compact_partition_params["source_partition_locator"] = params.get(
            "source_partition_locator"
        )
        compact_partition_params["primary_keys"] = params.get("primary_keys")
        compact_partition_params["rebase_source_partition_locator"] = params.get(
            "rebase_source_partition_locator"
        )
        compact_partition_params["rebase_source_partition_high_watermark"] = params.get(
            "rebase_source_partition_high_watermark"
        )
        compact_partition_params["hash_bucket_count"] = params.get("hash_bucket_count")
        compact_partition_params["deltacat_storage"] = params.get("deltacat_storage")
        compact_partition_params["compaction_artifact_s3_bucket"] = params.get(
            "compaction_artifact_s3_bucket"
        )
        compact_partition_params["properties"] = params.get("properties")
        compact_partition_params["compacted_file_content_type"] = params.get(
            "compacted_file_content_type"
        )
        compact_partition_params["list_deltas_kwargs"] = params.get(
            "list_deltas_kwargs"
        )
        compact_partition_params["pg_config"] = params.get("pg_config")
        compact_partition_params["read_kwargs_provider"] = params.get(
            "read_kwargs_provider"
        )
        compact_partition_params["s3_table_writer_kwargs"] = params.get(
            "s3_table_writer_kwargs"
        )
        return compact_partition_params

    @property
    def destination_partition_locator(self) -> Optional[dict]:
        return self["destination_partition_locator"]

    @property
    def last_stream_position_to_compact(self) -> Optional[int]:
        return self["last_stream_position_to_compact"]

    @property
    def source_partition_locator(self) -> Optional[dict]:
        return self["source_partition_locator"]

    @property
    def primary_keys(self) -> Optional[List[str]]:
        return list(self["primary_keys"])

    @property
    def rebase_source_partition_locator(self) -> Optional[dict]:
        return self["rebase_source_partition_locator"]

    @property
    def rebase_source_partition_high_watermark(self) -> Optional[int]:
        return self["rebase_source_partition_high_watermark"]

    @property
    def hash_bucket_count(self) -> Optional[int]:
        return self["hash_bucket_count"]

    @property
    def deltacat_storage(self) -> Optional[str]:
        return self["deltacat_storage"]

    @property
    def compaction_artifact_s3_bucket(self) -> Optional[str]:
        return self["compaction_artifact_s3_bucket"]

    @property
    def properties(self) -> Optional[Dict[str, str]]:
        return self["properties"]

    @property
    def compacted_file_content_type(self) -> Optional[ContentType]:
        return self["compacted_file_content_type"]

    @property
    def list_deltas_kwargs(self) -> Optional[dict]:
        return self["list_deltas_kwargs"]

    @property
    def pg_config(self) -> Optional[Any]:
        return self["pg_config"]

    @property
    def read_kwargs_provider(self) -> Optional[Any]:
        return self["read_kwargs_provider"]

    @property
    def s3_table_writer_kwargs(self) -> Optional[Any]:
        return self["s3_table_writer_kwargs"]

    @staticmethod
    def json_handler_for_compact_partition_params(obj):
        """
        A handler for the `json.dumps()` function that can be used to serialize sets to JSON.
        If the `set_default()` handler is passed as the `default` argument to the `json.dumps()` function, it will be called whenever a set object is encountered.
        The `set_default()` handler will then serialize the set as a list.
        """
        try:
            if isinstance(obj, set):
                return list(obj)
            elif hasattr(obj, "toJSON"):
                return obj.toJSON()
            else:
                return obj.__dict__
        except Exception:
            return obj.__class__.__name__

    def serialize(self) -> str:
        """
        Serializes itself to a json-formatted string

        Returns:
            The serialized object.

        """
        to_serialize: Dict[str, Any] = {}
        # individually try deepcopy the values from the self dictionary and just use the class name for the value when it is not possible to deepcopy
        for attr, value in self.items():
            try:
                to_serialize[attr] = copy.deepcopy(value)
            except Exception:  # if unable to deep copy the objects like module objects for example then just provide the class name at minimum
                to_serialize[attr] = value.__class__.__name__
        serialized_arguments_compact_partition_args: str = json.dumps(
            to_serialize,
            default=CompactPartitionParams.json_handler_for_compact_partition_params,
        )
        return serialized_arguments_compact_partition_args



import unittest
from unittest import mock
import sys


class TestGetCurrentClusterUtilization(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.ray_mock = mock.MagicMock()
        cls.ray_mock.cluster_resources.return_value = {
            "CPU": 10,
            "memory": 10,
            "object_store_memory": 5,
        }
        cls.ray_mock.available_resources.return_value = {
            "CPU": 6,
            "memory": 4,
            "object_store_memory": 5,
        }

        cls.module_patcher = mock.patch.dict("sys.modules", {"ray": cls.ray_mock})
        cls.module_patcher.start()

        # delete reference to reload from mocked ray
        if "deltacat.utils.resources" in sys.modules:
            del sys.modules["deltacat.utils.resources"]

        super().setUpClass()

    def test_sanity(self):
        from deltacat.utils.resources import ClusterUtilization

        result = ClusterUtilization.get_current_cluster_utilization()

        self.assertEqual(10, result.total_cpu)
        self.assertEqual(4, result.used_cpu)
        self.assertEqual(10, result.total_memory_bytes)
        self.assertEqual(5, result.total_object_store_memory_bytes)
        self.assertEqual(0, result.used_object_store_memory_bytes)
        self.assertEqual(6, result.used_memory_bytes)
        self.assertIsNotNone(result.used_resources)
